# -*- coding: utf-8 -*-
"""
Created on Sat Apr  3 15:42:40 2021

@author: veenstra

"""

import pytest
import os
import glob
import subprocess

# ACCEPTANCE TESTS VIA EXAMPLE SCRIPTS, these are the ones who are only meant to generate output files

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)
list_configfiles = glob.glob(os.path.join(dir_tests,'examples','*.py'))
dir_output_general = os.path.join(dir_tests,'examples_output')
os.makedirs(dir_output_general, exist_ok=True)

@pytest.mark.acceptance
@pytest.mark.parametrize("file_config", [pytest.param(file_config, id=os.path.basename(file_config).replace('.py','')) for file_config in list_configfiles])
def test_run_examples(file_config):
    # 1. Set up test data
    os.chdir(dir_output_general)

    p = subprocess.Popen(f"hatyan {file_config} --redirect-stdout --overwrite",
                          stderr=subprocess.STDOUT, # Merge stdout and stderr
                          stdout=subprocess.PIPE,
                          shell=True)
    # max 3 minutes per test, if it hangs longer the test is killed
    p.wait(180)
    
    if p.returncode:
        out, err = p.communicate()
        out_str = "\n".join([x.decode("utf-8") for x in out.split(b'\r\n')])
        raise RuntimeError(out_str)


# -*- coding: utf-8 -*-
"""
ddlpy_helpers.py contains functions to convert ddlpy timeseries dataframes to hatyan timeseries dataframes.
The package ddlpy is available at https://github.com/Deltares/ddlpy
"""

import pandas as pd

__all__ = ["ddlpy_to_hatyan"]


def ddlpy_to_hatyan(ddlpy_meas, ddlpy_meas_exttyp=None):
    """
    Convert ddlpy measurements to hatyan timeseries dataframe.

    Parameters
    ----------
    ddlpy_meas : pd.DataFrame
        ddlpy measurements dataframe. It is assumed that it contains numeric values that 
        represent waterlevel timeseries or waterlevel extremes (measured or astro).
    ddlpy_meas_exttyp : pd.DataFrame, optional
        ddlpy measurements dataframe. If it is supplied it is assumed that it contains 
        alfanumeric values and these represent tidal extreme types (high- and low waters).
        The default is None.

    Returns
    -------
    pd.DataFrame
        hatyan timeseries DataFrame with values/qualitycode/status columns. If ddlpy_meas_typ 
        is supplied, this DataFrame will also include a HWLWcode column.

    """
    
    
    ts_pd = ddlpy_to_hatyan_plain(ddlpy_meas, isnumeric=True)
    ts_pd['values'] /= 100 #convert from cm to m
    if ddlpy_meas_exttyp is None:
        return ts_pd
    
    # check if the contents of this dataframe is in deed extreme types
    assert len(ddlpy_meas) == len(ddlpy_meas_exttyp)
    typering_codes = ddlpy_meas_exttyp["Typering.Code"].drop_duplicates().values
    assert len(typering_codes) == 1
    assert typering_codes[0] == "GETETTPE"
    
    ts_pd_typ = ddlpy_to_hatyan_plain(ddlpy_meas_exttyp, isnumeric=False)
    
    ts_pd_combined = convert_exttype_str2num(ts_pd,ts_pd_typ)
    
    return ts_pd_combined
        

def ddlpy_to_hatyan_plain(ddlpy_meas, isnumeric=True):
    
    if ddlpy_meas.empty:
        raise ValueError("supplied dataframe is empty")
    
    cols_mustbe_unique = ['Grootheid.Code','Groepering.Code','Typering.Code','Hoedanigheid.Code']
    for col in cols_mustbe_unique:
        if len(ddlpy_meas[col].drop_duplicates()) != 1:
            raise Exception(f"ddlpy_meas['{col}'] is not unique")
    
    if isnumeric: 
        key_numericvalues = 'Meetwaarde.Waarde_Numeriek'
    else:
        # alfanumeric values for 'Typering.Code':'GETETTPE'
        key_numericvalues = 'Meetwaarde.Waarde_Alfanumeriek'
    
    ts_pd = pd.DataFrame({'values':ddlpy_meas[key_numericvalues],
                          'qualitycode':pd.to_numeric(ddlpy_meas['WaarnemingMetadata.KwaliteitswaardecodeLijst'],downcast='integer'),
                          'status':ddlpy_meas['WaarnemingMetadata.StatuswaardeLijst'].str[0],
                          })
    return ts_pd


def convert_exttype_str2num(ts_measwl_ext, ts_measwl_exttype):
    """
    TVL;1;1;hoogwater
    TVL;1;2;laagwater
    TVL;1;3;laagwater 1
    TVL;1;4;topagger
    TVL;1;5;laagwater 2
    """
    
    ts_measwl_ext.loc[ts_measwl_exttype['values']=='hoogwater','HWLWcode'] = 1
    ts_measwl_ext.loc[ts_measwl_exttype['values']=='laagwater','HWLWcode'] = 2
    ts_measwl_ext.loc[ts_measwl_exttype['values']=='laagwater 1','HWLWcode'] = 3
    ts_measwl_ext.loc[ts_measwl_exttype['values']=='topagger','HWLWcode'] = 4
    ts_measwl_ext.loc[ts_measwl_exttype['values']=='laagwater 2','HWLWcode'] = 5
    ts_measwl_ext['HWLWcode'] = ts_measwl_ext['HWLWcode'].astype(int)
    return ts_measwl_ext




# -*- coding: utf-8 -*-
"""
Created on Mon May  6 16:11:38 2024

@author: veenstra
"""

import pytest
import os
import numpy as np
import pandas as pd
import hatyan
import pytz
import datetime as dt
from hatyan.analysis_prediction import MatrixConditionTooHigh

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)
dir_testdata = os.path.join(dir_tests,'data_unitsystemtests')


@pytest.mark.unittest
def test_analysis_settings():
    
    current_station = 'VLISSGN'
    
    #comp0
    file_data_comp0 = [os.path.join(dir_testdata,'%s_obs%i.txt'%(current_station, file_id)) for file_id in [1]]
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    
    ts_comp_nfac1_fualltimes0_xfac1_peryear0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, analysis_perperiod=False)
    
    ts_comp_onecomp = hatyan.analysis(ts=ts_measurements_group0, const_list=['M2'], nodalfactors=True, fu_alltimes=True, xfac=True)
    
    ts_comp_nfac1_fualltimes1_xfac1 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=True)
    ts_comp_nfac1_fualltimes0_xfac1 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True)
    ts_comp_nfac1_fualltimes1_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=False)
    ts_comp_nfac1_fualltimes0_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=False)
    ts_comp_nfac0_fualltimes0_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=False, fu_alltimes=False, xfac=False)
    
    comp_A_expected = np.array([9.92687566e-04, 3.11039747e-02, 9.73652407e-02, 6.68645368e-02,
                                4.02714135e-02, 2.44247279e-02, 1.34401343e-01, 2.69848016e-01,
                                1.74658516e+00, 1.33037114e-01, 4.78882284e-01, 4.42634321e-02,
                                1.93477077e-02, 4.29425377e-02, 1.28103067e-01, 8.89165480e-02,
                                4.66780541e-02, 8.50788597e-02, 9.01287840e-02, 3.06546626e-02,
                                4.72786278e-02, 1.47588322e-02])
    
    comp_phi_expected = np.array([  0.        , 141.78046951, 188.22394202,   7.35166063,
                                  279.19176166, 122.65613222, 159.77053294,  33.54516691,
                                   59.02752661, 256.04742485, 117.45538945, 345.62388999,
                                  200.90891452,  92.80114988, 115.31391482, 176.58434752,
                                   77.32631928, 103.4830287 , 153.94793646, 110.76924584,
                                  157.3933541 , 220.41306556])
    
    assert (np.abs(ts_comp_nfac1_fualltimes0_xfac1_peryear0-ts_comp_nfac1_fualltimes0_xfac1) < 10E-9).all().all()
    assert (np.abs(ts_comp_nfac1_fualltimes0_xfac1_peryear0['phi_deg'].values-comp_phi_expected) < 10E-9).all()
    assert (np.abs(ts_comp_nfac1_fualltimes0_xfac1_peryear0['A'].values-comp_A_expected) < 10E-9).all()

    assert np.allclose(ts_comp_nfac1_fualltimes0_xfac1_peryear0.loc["M2"].values, np.array([ 1.74658516, 59.02752661]))
    assert np.allclose(ts_comp_onecomp.loc["M2"].values, np.array([ 1.74229247, 58.77828766]))
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1.loc["M2"].values, np.array([ 1.74649803, 59.01827634]))
    assert np.allclose(ts_comp_nfac1_fualltimes0_xfac1.loc["M2"].values, np.array([ 1.74658516, 59.02752661]))
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac0.loc["M2"].values, np.array([ 1.7623892 , 59.01684785]))
    assert np.allclose(ts_comp_nfac1_fualltimes0_xfac0.loc["M2"].values, np.array([ 1.76257914, 59.02752661]))
    assert np.allclose(ts_comp_nfac0_fualltimes0_xfac0.loc["M2"].values, np.array([ 1.72889408, 57.192191  ]))


@pytest.mark.unittest
def test_analysis_settings_extended():
    
    current_station = 'VLISSGN'
    
    #comp0
    file_data_comp0 = [os.path.join(dir_testdata,'%s_obs%i.txt'%(current_station, file_id)) for file_id in [1,2]]
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    
    ts_comp_nfac1_fualltimes1_xfac1_peryear0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=True, analysis_perperiod=False)
    
    ts_comp_nfac1_fualltimes1_xfac1_permonth0, comp_allperiods = hatyan.analysis(ts=ts_measurements_group0, const_list='month', 
                                                                                 nodalfactors=True, fu_alltimes=True, xfac=True, 
                                                                                 analysis_perperiod='M', return_allperiods=True)
    assert comp_allperiods.shape == (22, 48)
    
    ts_comp_nfac1_fualltimes1_xfac1 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=True, analysis_perperiod='Y')
    ts_comp_nfac1_fualltimes0_xfac1 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, analysis_perperiod='Y')
    ts_comp_nfac1_fualltimes1_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=False, analysis_perperiod='Y')
    ts_comp_nfac1_fualltimes0_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=False, analysis_perperiod='Y')
    ts_comp_nfac0_fualltimes0_xfac0 = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=False, fu_alltimes=False, xfac=False, analysis_perperiod='Y')
    
    # assert if the results are somewhat close to each other
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1_peryear0, ts_comp_nfac1_fualltimes1_xfac1_permonth0, rtol=1e-1, atol=1e-1)
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1, ts_comp_nfac1_fualltimes0_xfac1, rtol=1e-2, atol=1e-3)
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1, ts_comp_nfac1_fualltimes0_xfac1, rtol=1e-2, atol=1e-2)
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1, ts_comp_nfac1_fualltimes1_xfac0, rtol=1e-2, atol=1e-2)
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1, ts_comp_nfac1_fualltimes0_xfac0, rtol=1e-2, atol=1e-2)
    assert np.allclose(ts_comp_nfac1_fualltimes1_xfac1, ts_comp_nfac0_fualltimes0_xfac0, rtol=1e1, atol=1e0)


@pytest.mark.unittest
def test_analysis_settings_perperiod():
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    comp_mean, comp_all = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, 
                                          analysis_perperiod="M", return_allperiods=True)
    assert len(comp_mean.attrs) > 0
    assert len(comp_all.attrs) > 0


@pytest.mark.unittest
def test_analysis_component_splitting():
    current_station = 'D15'
    file_ts = os.path.join(dir_testdata,'%s_obs1.txt'%(current_station))
    cs_comps = pd.DataFrame({'CS_comps_from':['K1','N2','2MN2','S2','S2'],
                             'CS_comps_derive':['P1','NU2','LABDA2','K2','T2'],
                             'CS_ampfacs':[0.33,0.22,0.48,0.29,0.05],
                             'CS_degincrs':[-11,-24,174,1,-24]})
    
    ts_meas = hatyan.read_dia(filename=file_ts, station=current_station)
    comp_cs_atonce = hatyan.analysis(ts=ts_meas, const_list='month', 
                                     analysis_perperiod=False, cs_comps=cs_comps)
    comp_cs_peryear = hatyan.analysis(ts=ts_meas, const_list='month', 
                                      analysis_perperiod="Y", cs_comps=cs_comps)
    comp_cs_permonth = hatyan.analysis(ts=ts_meas, const_list='month', 
                                       analysis_perperiod="M", cs_comps=cs_comps)
    
    assert np.allclose(comp_cs_atonce.loc["M2"].values, [  0.6222659 , 187.24099172])
    assert np.allclose(comp_cs_peryear.loc["M2"].values, [  0.6222659 , 187.24099172])
    assert np.allclose(comp_cs_permonth.loc["M2"].values, [  0.6222659 , 187.24099172])


@pytest.mark.unittest
def test_analysis_foreman():
    current_station = 'VLISSGN'
    
    #comp0
    file_data_comp0 = [os.path.join(dir_testdata,'%s_obs%i.txt'%(current_station, file_id)) for file_id in [1]]
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0, station=current_station)
    
    comp_frommeas_schu = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=True, source='schureman')
    comp_frommeas_for = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=True, xfac=True, source='foreman')
    schu_comp_phi_expected =  np.array([  0.        , 141.68386391, 188.27101759,   7.3992028 ,
                                       279.17807514, 122.81257343, 159.77061119,  33.51788732,
                                        59.01827634, 256.05796043, 117.47528038, 345.63438447,
                                       201.05307505,  92.81986396, 115.2820895 , 176.56952991,
                                        77.30925229, 103.43837741, 153.90952097, 110.64017054,
                                       157.30038107, 220.24021917])
    schu_comp_A_expected =    np.array([9.93677030e-04, 3.11102518e-02, 9.74940999e-02, 6.69089179e-02,
                                       4.01204828e-02, 2.43192049e-02, 1.34386805e-01, 2.69878623e-01,
                                       1.74649803e+00, 1.32976762e-01, 4.79019453e-01, 4.42086557e-02,
                                       1.92695651e-02, 4.29786476e-02, 1.28078709e-01, 8.89513755e-02,
                                       4.67055728e-02, 8.50832246e-02, 9.01742533e-02, 3.06526796e-02,
                                       4.72907179e-02, 1.47366661e-02])
    for_comp_phi_expected =  np.array([  0.        , 321.7403689 ,   8.15729584, 187.40999172,
                                       279.42669414, 122.65017113, 160.05967765,  33.32750808,
                                        59.01572119, 256.18495318, 117.35826025, 345.41402817,
                                       201.17575979,  92.60563462, 115.27544986, 176.43679913,
                                        77.08771066, 103.42351654, 153.76456299, 110.600143  ,
                                       157.16101443, 220.13261581])
    for_comp_A_expected =    np.array([9.96058525e-04, 3.01459094e-02, 9.72007804e-02, 6.69287134e-02,
                                       3.99816712e-02, 2.42471615e-02, 1.33778086e-01, 2.75247289e-01,
                                       1.76155491e+00, 1.39007149e-01, 4.85825897e-01, 4.40858825e-02,
                                       1.92198515e-02, 4.29817009e-02, 1.29452749e-01, 9.04975848e-02,
                                       4.66816274e-02, 8.62181973e-02, 9.27486375e-02, 3.13117938e-02,
                                       4.82570962e-02, 1.46874202e-02])
    
    assert (np.abs(comp_frommeas_schu['phi_deg'].values-schu_comp_phi_expected) < 10E-9).all()
    assert (np.abs(comp_frommeas_schu['A'].values-schu_comp_A_expected) < 10E-9).all()
    assert (np.abs(comp_frommeas_for['phi_deg'].values-for_comp_phi_expected) < 10E-9).all()
    assert (np.abs(comp_frommeas_for['A'].values-for_comp_A_expected) < 10E-9).all()


@pytest.mark.unittest
def test_analysis_settings_timeseries_tooshort():
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    
    with pytest.raises(MatrixConditionTooHigh) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list="all_schureman_originalorder")
    assert "condition of xTx matrix is too high" in str(e.value)

    with pytest.raises(ValueError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list="year", analysis_perperiod="M")
    assert "all nans" in str(e.value)
    
    with pytest.raises(ValueError) as e:
        _ = hatyan.analysis(ts=ts_pd.iloc[:1], const_list="year")
    assert "provided timeseries is less than 2 timesteps long" in str(e.value)
    
    ts_pd_manynans = ts_pd.copy()
    ts_pd_manynans.iloc[1:] = np.nan
    with pytest.raises(ValueError) as e:
        _ = hatyan.analysis(ts=ts_pd_manynans, const_list="year")
    assert "provided timeseries is less than 2 timesteps long" in str(e.value)

    ts_pd_manynans = ts_pd.copy()
    ts_pd_manynans.iloc[2:] = np.nan
    with pytest.raises(MatrixConditionTooHigh) as e:
        _ = hatyan.analysis(ts=ts_pd_manynans, const_list="year")
    assert "condition of xTx matrix is too high" in str(e.value)


@pytest.mark.unittest
def test_analysis_invalid_timeseries_index():
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    
    ts_pd_wrongindex = ts_pd.reset_index()
    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd_wrongindex, const_list="year")
    assert "ts.index is not of expected type" in str(e.value)
    
    ts_pd_duplicated = pd.concat([ts_pd,ts_pd])
    with pytest.raises(ValueError) as e:
        _ = hatyan.analysis(ts=ts_pd_duplicated, const_list="year")
    assert "duplicate timesteps in provided timeseries" in str(e.value)


@pytest.mark.unittest
def test_analysis_settings_invalid_values():
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    
    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], nodalfactors=1)
    assert str(e.value) == "invalid nodalfactors=1 type, should be bool"

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], xfac=1)
    assert str(e.value) == "invalid xfac=1 type, should be bool or dict"

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], fu_alltimes=1)
    assert str(e.value) == "invalid fu_alltimes=1 type, should be bool"

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], source=1)
    assert str(e.value) == "invalid source=1 type, should be str"

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], source="aa")
    assert str(e.value) == 'invalid source aa, should be "schureman" or "foreman"'

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], analysis_perperiod="aa")
    assert str(e.value) == 'invalid analysis_perperiod=aa type, should be False or Y/Q/M'

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], return_allperiods=1)
    assert str(e.value) == 'invalid return_allperiods=1 type, should be bool'

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], cs_comps=1)
    assert str(e.value) == "invalid cs_comps type, should be dict"

    with pytest.raises(KeyError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], cs_comps={})
    assert str(e.value) == "'cs_comps does not contain CS_comps_derive'"

    with pytest.raises(TypeError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2"], max_matrix_condition="aa")
    assert str(e.value) == "invalid aa type, should be int or float"


@pytest.mark.unittest
def test_analysis_invalid_constituent_lists():
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    
    with pytest.raises(KeyError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list="M2")
    assert 'listtype "M2" is not implemented in hatyan.get_const_list_hatyan' in str(e.value)

    with pytest.raises(ValueError) as e:
        _ = hatyan.analysis(ts=ts_pd, const_list=["M2","M2"])
    assert "remove duplicate constituents from const_list" in str(e.value)


@pytest.mark.unittest
def test_analysis_component_splitting_invalid_components():
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    cs_comps_valid = pd.DataFrame({'CS_comps_derive':['P1','NU2','LABDA2','K2','T2'],
                             'CS_comps_from':['K1','N2','2MN2','S2','S2'],
                             'CS_ampfacs':[0.36,0.38,0.44,0.30,0.07],
                             'CS_degincrs':[5,-22,180,3,-22]})
    with pytest.raises(ValueError) as e:
        _= hatyan.analysis(ts=ts_pd, const_list="year", cs_comps=cs_comps_valid)
    assert str(e.value) == "component P1 requested via component splitting, but already present in const_list"


@pytest.mark.unittest
def test_analysis_deprecatedsettings():
    times_pred = slice(dt.datetime(2009,12,31,14),dt.datetime(2010,1,2,12), "1min")
    file_data_comp0 = os.path.join(dir_testdata,'DENHDR_ana.txt')
    comp = hatyan.read_components(filename=file_data_comp0)
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.analysis(comp=comp, times=times_pred, CS_comps=1)
    assert str(e.value) == "Argument 'CS_comps' has been deprecated for hatyan.analysis(), use 'cs_comps' instead"

    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.analysis(comp=comp, times=times_pred, xTxmat_condition_max=1)
    assert str(e.value) == "Argument 'xTxmat_condition_max' has been deprecated for hatyan.analysis(), use 'max_matrix_condition' instead"


@pytest.mark.unittest
def test_prediction_settings():
    times_pred = slice(dt.datetime(2009,12,31,14),dt.datetime(2010,1,2,12), "1min")
    current_station = 'DENHDR'
    
    file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    COMP_merged = hatyan.read_components(filename=file_data_comp0)
    COMP_merged.attrs["nodalfactors"] = True
    COMP_merged.attrs["fu_alltimes"] = False
    COMP_merged.attrs["xfac"] = True
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.55761625, -0.55022362, -0.54263955, -0.53486334, -0.52689445,
       -0.51873254, -0.51037744, -0.50182919, -0.49308803, -0.48415438])
    assert np.allclose(ts_pred["values"].values[:10], expected)
    
    COMP_merged.attrs["nodalfactors"] = True
    COMP_merged.attrs["fu_alltimes"] = True
    COMP_merged.attrs["xfac"] = True
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.55758197, -0.55018925, -0.54260512, -0.53482887, -0.52685996,
           -0.51869805, -0.51034297, -0.50179476, -0.49305366, -0.4841201 ])
    assert np.allclose(ts_pred["values"].values[:10], expected)
    
    COMP_merged.attrs["nodalfactors"] = True
    COMP_merged.attrs["fu_alltimes"] = True
    COMP_merged.attrs["xfac"] = {"M2":0.8}
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.55220418, -0.54482864, -0.53726402, -0.52950961, -0.52156487,
           -0.51342947, -0.50510324, -0.49658621, -0.48787861, -0.47898087])
    assert np.allclose(ts_pred["values"].values[:10], expected)
    
    COMP_merged.attrs["nodalfactors"] = True
    COMP_merged.attrs["fu_alltimes"] = True
    COMP_merged.attrs["xfac"] = False
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.55115935, -0.54379567, -0.53624298, -0.52850057, -0.52056791,
           -0.51244465, -0.50413063, -0.49562589, -0.48693065, -0.47804533])
    assert np.allclose(ts_pred["values"].values[:10], expected)
    
    COMP_merged.attrs["nodalfactors"] = True
    COMP_merged.attrs["fu_alltimes"] = False
    COMP_merged.attrs["xfac"] = False
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.55120816, -0.54384448, -0.53629177, -0.52854932, -0.52061659,
           -0.51249324, -0.5041791 , -0.49567421, -0.4869788 , -0.47809328])
    assert np.allclose(ts_pred["values"].values[:10], expected)
    
    COMP_merged.attrs["nodalfactors"] = False
    COMP_merged.attrs["fu_alltimes"] = False
    COMP_merged.attrs["xfac"] = False
    ts_pred = hatyan.prediction(comp=COMP_merged, times=times_pred)
    expected = np.array([-0.61931214, -0.61234662, -0.60517618, -0.59779951, -0.59021548,
           -0.58242314, -0.57442171, -0.56621065, -0.55778958, -0.54915834])
    assert np.allclose(ts_pred["values"].values[:10], expected)


@pytest.mark.unittest
def test_prediction_settings_invalid_values():
    current_station = 'VLISSGN'
    file_comp = os.path.join(dir_testdata, f'{current_station}_ana.txt')
    
    # missing xfac attribute
    comp = hatyan.read_components(filename=file_comp)
    comp.attrs.pop("xfac")
    with pytest.raises(KeyError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == "'xfac'"
    
    # no settings attrs at all
    comp = hatyan.read_components(filename=file_comp)
    comp.attrs = {}
    with pytest.raises(KeyError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == "'nodalfactors'"

    # incorrect settings
    comp = hatyan.read_components(filename=file_comp)
    comp.attrs["nodalfactors"] = 1
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == "invalid nodalfactors=1 type, should be bool"

    comp = hatyan.read_components(filename=file_comp)
    comp.attrs["fu_alltimes"] = 1
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == "invalid fu_alltimes=1 type, should be bool"

    comp = hatyan.read_components(filename=file_comp)
    comp.attrs["xfac"] = 1
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == "invalid xfac=1 type, should be bool or dict"

    comp = hatyan.read_components(filename=file_comp)
    comp.attrs["source"] = "aa"
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp)
    assert str(e.value) == 'invalid source aa, should be "schureman" or "foreman"'


@pytest.mark.unittest
def test_prediction_settings_invalid_times():
    current_station = 'VLISSGN'
    file_comp = os.path.join(dir_testdata, f'{current_station}_ana.txt')
    comp = hatyan.read_components(filename=file_comp)
    
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp, times=1)
    assert "times argument can be of type" in str(e.value)
    
    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp, timestep=1)
    assert str(e.value) == "prediction() atonce, so 'timestep' argument not allowed"


@pytest.mark.unittest
def test_prediction_perperiod_settings_invalid_timestep():
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    _, comp_all = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, 
                                          analysis_perperiod="M", return_allperiods=True)
    
    with pytest.raises(ValueError) as e:
        hatyan.prediction(comp=comp_all, timestep=60)
    assert str(e.value) == "Invalid frequency: 60"

    with pytest.raises(TypeError) as e:
        hatyan.prediction(comp=comp_all, timestep=60, times=60)
    assert "prediction() per period, so 'times' argument not allowed" == str(e.value)


@pytest.mark.unittest
def test_analysis_1018():
    """
    this tests checks if we can do an analysis with timesteps that are outofbounds 
    in pandas when using unit="ns" (nanoseconds). 
    We replace the times from a dataframe with these timesteps, but with unit="us" (microseconds). 
    The analysis succeeds with pandas versions that support the unit argument, so all versions that hatyan supports.
    However, if a diafile would contain 1018 timestamps, the process will fail since 
    the dia reader will still read it as nanoseconds. This might also be the case 
    when writing the prediction to a diafile.
    """
    file_dia = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_pd = hatyan.read_dia(file_dia)
    
    # convert to year 1018
    ts_pd.index = pd.date_range('1018-01-01 00:00:00+01:00', '1018-12-31 23:00:00+01:00', freq="60min", unit="us")
    
    #prediction
    comp = hatyan.analysis(ts=ts_pd, const_list="year")
    
    assert np.allclose(comp.loc["M2"].values, [ 1.78668261, 45.85671448])


@pytest.mark.unittest
def test_prediction_1018():
    current_station = 'DENHDR'
    
    times_pred = slice(dt.datetime(1018,7,21),dt.datetime(1018,7,21,3), "10min")
    
    file_data_comp0 = os.path.join(dir_testdata,'%s_ana.txt'%(current_station))
    COMP_merged = hatyan.read_components(filename=file_data_comp0)
    
    #prediction
    ts_prediction = hatyan.prediction(comp=COMP_merged, times=times_pred)
    
    ts_prediction_times = np.array([dt.datetime(1018, 7, 21, 0, 0),
                                    dt.datetime(1018, 7, 21, 0, 10),
                                    dt.datetime(1018, 7, 21, 0, 20),
                                    dt.datetime(1018, 7, 21, 0, 30),
                                    dt.datetime(1018, 7, 21, 0, 40),
                                    dt.datetime(1018, 7, 21, 0, 50),
                                    dt.datetime(1018, 7, 21, 1, 0),
                                    dt.datetime(1018, 7, 21, 1, 10),
                                    dt.datetime(1018, 7, 21, 1, 20),
                                    dt.datetime(1018, 7, 21, 1, 30),
                                    dt.datetime(1018, 7, 21, 1, 40),
                                    dt.datetime(1018, 7, 21, 1, 50),
                                    dt.datetime(1018, 7, 21, 2, 0),
                                    dt.datetime(1018, 7, 21, 2, 10),
                                    dt.datetime(1018, 7, 21, 2, 20),
                                    dt.datetime(1018, 7, 21, 2, 30),
                                    dt.datetime(1018, 7, 21, 2, 40),
                                    dt.datetime(1018, 7, 21, 2, 50),
                                    dt.datetime(1018, 7, 21, 3, 0)], dtype=np.datetime64)
    ts_prediction_times_pd = pd.DatetimeIndex(ts_prediction_times).tz_localize("UTC+01:00")
    ts_prediction_vals = np.array([-0.77416669, -0.78821831, -0.79718123, -0.80126649, -0.80055319,
                                   -0.79466499, -0.78252257, -0.76226094, -0.73136835, -0.68705389,
                                   -0.62679467, -0.54896408, -0.4534113 , -0.34185962, -0.218017  ,
                                   -0.08734459,  0.04350389,  0.16749106,  0.27811846])
    
    assert (np.abs(ts_prediction['values'].values-ts_prediction_vals) < 10E-9).all()
    assert (np.abs(ts_prediction.index-ts_prediction_times_pd) < dt.timedelta(days=10E-9)).all()


@pytest.mark.unittest
def test_prediction_comp_and_times_different_timezones():
    """
    it is possible to supply a timezone via times, but the comp dataframe also contains a timezone already.
    The components timezone is leading, but the times will be converted to that timezone also.
    From the below test, both predictions are therefore UTC+01:00, but the times are shifted
    https://github.com/Deltares/hatyan/issues/334
    """
    
    current_station = 'VLISSGN'
    
    const_list = hatyan.get_const_list_hatyan('year') # 94 constituents
    
    file_comp = os.path.join(dir_testdata,f'{current_station}_obs1.txt')
    ts_meas = hatyan.read_dia(filename=file_comp, station=current_station)
    
    comp_naive = hatyan.analysis(ts=ts_meas, const_list=const_list)
    comp_naive.attrs["tzone"] = None
    comp_met = hatyan.analysis(ts=ts_meas, const_list=const_list)
    
    times_naive = slice("2019-01-01","2020-01-01", "10min")
    times_met = pd.date_range("2019-01-01","2020-01-01", freq="10min", tz="UTC+01:00")
    times_utc = pd.date_range("2019-01-01","2020-01-01", freq="10min", tz="UTC+00:00")
    
    pred_naive = hatyan.prediction(comp=comp_naive, times=times_naive)
    pred_met = hatyan.prediction(comp=comp_met, times=times_met)
    pred_utc = hatyan.prediction(comp=comp_met, times=times_utc)
    
    assert pred_naive.index.tz is None
    assert pred_naive.index.freq is not None
    assert pred_naive.index[0] == pd.Timestamp('2019-01-01 00:00:00')
    assert pred_met.index.tz == dt.timezone(dt.timedelta(seconds=3600))
    assert pred_met.index.freq is not None
    assert pred_met.index[0] == pd.Timestamp('2019-01-01 00:00:00+0100')
    assert pred_utc.index.tz == dt.timezone.utc
    assert pred_utc.index.freq is not None
    assert pred_utc.index[0] == pd.Timestamp('2019-01-01 00:00:00+0000')
    assert ((pred_naive - pred_met.tz_localize(None)).dropna()["values"] < 1e-9).all()
    assert ((pred_utc - pred_met).dropna()["values"] < 1e-9).all()


@pytest.mark.unittest
def test_prediction_times_tznaive_comp_tzaware(caplog):
    """
    https://github.com/Deltares/hatyan/issues/334
    """
    comp = pd.DataFrame({"A": [1, 0.5, 0.2],
                         "phi_deg": [10,15,20]}, 
                        index=["M2","M4","S2"])
    comp.attrs["nodalfactors"] = True
    comp.attrs["fu_alltimes"] = True
    comp.attrs["xfac"] = False
    comp.attrs["source"] = "schureman"
    comp.attrs["tzone"] = "UTC+01:00"
    dtindex = pd.date_range("2020-01-01","2020-01-02", freq="10min")
    hatyan.prediction(comp, times=dtindex)
    warning_text = ("provided times are timezone-naive and provided components are "
                    "timezone-aware. The times are being interpreted as if they would "
                    "have the same timezone as the components: UTC+01:00")
    assert warning_text in caplog.text


@pytest.mark.unittest
def test_prediction_raise_mixed_tznaive_tzaware():
    """
    https://github.com/Deltares/hatyan/issues/334
    """
    comp = pd.DataFrame({"A": [1, 0.5, 0.2],
                         "phi_deg": [10,15,20]}, 
                        index=["M2","M4","S2"])
    comp.attrs["nodalfactors"] = True
    comp.attrs["fu_alltimes"] = True
    comp.attrs["xfac"] = False
    comp.attrs["source"] = "schureman"
    comp.attrs["tzone"] = None
    dtindex = pd.date_range("2020-01-01 00:00 +00:00","2020-01-02 00:00 +00:00", freq="10min")
    with pytest.raises(ValueError) as e:
        hatyan.prediction(comp, times=dtindex)
    assert "provided times and components should both be timezone-aware or timezone-naive, not mixed." in str(e.value)


@pytest.mark.unittest
def test_prediction_comp_no_timezone():
    """
    https://github.com/Deltares/hatyan/issues/317
    """
    comp = pd.DataFrame({"A": [1, 0.5, 0.2],
                         "phi_deg": [10,15,20]}, 
                        index=["M2","M4","S2"])
    comp.attrs["nodalfactors"] = True
    comp.attrs["fu_alltimes"] = True
    comp.attrs["xfac"] = False
    comp.attrs["source"] = "schureman"
    dtindex = pd.date_range("2020-01-01","2020-01-02", freq="10min")
    pred = hatyan.prediction(comp, times=dtindex)
    assert pred.index.tz is None
    assert pred.index[0] == pd.Timestamp('2020-01-01 00:00:00')
    assert pred.index[-1] == pd.Timestamp('2020-01-02 00:00:00')


@pytest.mark.unittest
def test_prediction_perperiod_month():
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    comp_mean, comp_all = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, 
                                          analysis_perperiod="M", return_allperiods=True)
    
    ts_pred_atonce = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index)
    
    ts_pred_allmonths = hatyan.prediction(comp=comp_all, timestep="60min")
    
    expected_atonce = np.array([-1.39664945, -0.85846188, -0.22419191,  0.7596607 ,  1.91604743,
            2.17294986,  1.57508019,  0.88085591, -0.01977715, -1.01827975])
    expected_allmonths = np.array([-1.51643013, -1.06743931, -0.51656439,  0.40987597,  1.66862452,
            2.09709427,  1.60408414,  0.9891484 ,  0.13523824, -0.89396145])
    
    assert np.allclose(ts_pred_atonce["values"].values[:10], expected_atonce)
    assert np.allclose(ts_pred_allmonths["values"].values[:10], expected_allmonths)
    assert len(ts_pred_atonce) == len(ts_pred_allmonths)
    assert ts_pred_allmonths.index.tz == pytz.FixedOffset(60)


@pytest.mark.unittest
def test_prediction_perperiod_year():
    # TODO: when the prediction function does not treat Y and M differently, this testcase can be dropped
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    comp_mean, comp_all = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, 
                                          analysis_perperiod="Y", return_allperiods=True)
    
    ts_pred_atonce = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index)
    
    ts_pred_allyears = hatyan.prediction(comp=comp_all, timestep="60min")
    
    expected_atonce = np.array([-1.38346859, -0.84555223, -0.21307598,  0.76601685,  1.91438215,
            2.16373654,  1.56241672,  0.86686561, -0.0329582 , -1.02668776])
    expected_allyears = np.array([-1.38346859, -0.84555223, -0.21307598,  0.76601685,  1.91438215,
            2.16373654,  1.56241672,  0.86686561, -0.0329582 , -1.02668776])
    
    assert np.allclose(ts_pred_atonce["values"].values[:10], expected_atonce)
    assert np.allclose(ts_pred_allyears["values"].values[:10], expected_allyears)
    assert len(ts_pred_atonce) == len(ts_pred_allyears)
    assert ts_pred_allyears.index.tz == pytz.FixedOffset(60)


@pytest.mark.unittest
def test_prediction_hasrequiredargs():
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    comp_mean, comp_all = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True, 
                                          analysis_perperiod="M", return_allperiods=True)
    
    with pytest.raises(TypeError) as e:
        _ = hatyan.prediction(comp=comp_mean)
    assert str(e.value) == "prediction() atonce, so 'times' argument should not be None"
    
    with pytest.raises(TypeError) as e:
        _ = hatyan.prediction(comp=comp_all)
    assert str(e.value) == "prediction() per period, so 'timestep' argument should not be None"


@pytest.mark.unittest
def test_prediction_deprecatedsettings():
    file_data_comp0 = os.path.join(dir_testdata,'VLISSGN_obs1.txt')
    ts_measurements_group0 = hatyan.read_dia(filename=file_data_comp0)
    
    comp_mean = hatyan.analysis(ts=ts_measurements_group0, const_list='month', nodalfactors=True, fu_alltimes=False, xfac=True)
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, times_pred_all="")
    assert str(e.value) == "Argument 'times_pred_all' has been deprecated for hatyan.prediction(), use 'times' instead"
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, times_ext="")
    assert str(e.value) == "Argument 'times_ext' has been deprecated for hatyan.prediction(), use 'times' instead"
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, nodalfactors=False)
    assert "as attribute of the component dataframe" in str(e.value)
        
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, xfac=False)
    assert "as attribute of the component dataframe" in str(e.value)
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, fu_alltimes=False)
    assert "as attribute of the component dataframe" in str(e.value)
    
    with pytest.raises(DeprecationWarning) as e:
        _ = hatyan.prediction(comp=comp_mean, times=ts_measurements_group0.index, source=False)
    assert "as attribute of the component dataframe" in str(e.value)



# -*- coding: utf-8 -*-
"""
foreman.py contains all foreman definitions now embedded in hatyan. The dataset is derived from "M.G.G. Foreman (2004), Manual for Tidal Heights Analysis and Prediction, Institute of Ocean Sciences (Patricia Bay, Sidney B.C. Canada)"
"""

import os
import pandas as pd
import numpy as np
import functools
import datetime as dt

file_path = os.path.realpath(__file__)


#################################################
################# FILECONTENTS ##################
#################################################


@functools.lru_cache() #only caching this already makes foreman slightly faster
def get_foreman_doodson_nodal_harmonic(lat_deg=51.45):
    """
    Omzetten van het tweede deel van de foremantabel in een pandas DataFrame met harmonische (+satellite) componenten.

    Parameters
    ----------
    const_list : TYPE
        DESCRIPTION.
    lat_deg : TYPE, optional
         0 in degrees from equator (pos N, neg S). For R1 and R2, maar variatie heeft niet erg veel invloed op A en phi. The default is 51.45.

    Returns
    -------
    foreman_harmonic_doodson_all : TYPE
        DESCRIPTION.
    foreman_harmonic_nodal_all : TYPE
        DESCRIPTION.
    foreman_harmonic_doodson : TYPE
        DESCRIPTION.
    foreman_harmonic_nodal : TYPE
        DESCRIPTION.

    """
    
    foreman_file = os.path.join(os.path.dirname(file_path),'data','data_foreman_harmonic.txt')
    foreman_harmonic_raw = pd.read_csv(foreman_file, comment='#', names=range(16), index_col=0, skip_blank_lines=True, sep="\\s+")
    bool_dupl_index = foreman_harmonic_raw.index.duplicated(keep='first')
    
    lat_rad = np.deg2rad(lat_deg)
    
    R1 = 0.36309*(1.-5.*np.sin(lat_rad)*np.sin(lat_rad))/np.sin(lat_rad) #-1 #for lat=50N
    R2 = 2.59808*np.sin(lat_rad) #2 #for lat=50N
    
    #get all forman harmonic doodson values (first occurences of duplicated component names)
    foreman_doodson_harmonic_lun = foreman_harmonic_raw.loc[~bool_dupl_index].dropna(axis=1)
    foreman_doodson_harmonic_lun[5] = foreman_doodson_harmonic_lun[5].astype(float) #for some reason, column 5/N remains dtype object
    if not len(foreman_doodson_harmonic_lun.columns) == 8:
        raise Exception('unexpected amount of columns for harmonics, inconsistent file')
    foreman_doodson_harmonic_lun.columns = ['T','S','H','P','N','P1','EDN','nsats']
    foreman_doodson_harmonic_lun.index.name = None
    
    #convert from lunar to solar and reverse automatic column sorting
    omega1 = foreman_doodson_harmonic_lun.loc[:,'T']
    corr_array = pd.DataFrame({'S':-omega1,'H':omega1})
    foreman_doodson_harmonic = foreman_doodson_harmonic_lun.add(corr_array,fill_value=0)
    foreman_doodson_harmonic = foreman_doodson_harmonic[foreman_doodson_harmonic_lun.columns]
    
    #get all foreman harmonic nodal values (non-first occurences of duplicated component names)
    foreman_nodal_harmonic_wide = foreman_harmonic_raw.loc[bool_dupl_index]
    #reshape from 15 to 5 columns and drop nan lines
    nrows_wide = len(foreman_nodal_harmonic_wide)
    foreman_nodal_harmonic = pd.DataFrame(foreman_nodal_harmonic_wide.values.reshape(nrows_wide*3,5),
                                          index=foreman_nodal_harmonic_wide.index.repeat(3),
                                          columns=['P','N','P1','EDN','factor']).dropna(axis=0)
    
    #multiply with R1/R2 if applicable and convert entire dataframe to floats
    bool_R1 = foreman_nodal_harmonic['factor'].str.contains('R1')
    bool_R2 = foreman_nodal_harmonic['factor'].str.contains('R2')
    foreman_nodal_harmonic.loc[bool_R1,'factor'] = foreman_nodal_harmonic.loc[bool_R1,'factor'].str.replace('R1','').astype(float)*R1
    foreman_nodal_harmonic.loc[bool_R2,'factor'] = foreman_nodal_harmonic.loc[bool_R2,'factor'].str.replace('R2','').astype(float)*R2
    foreman_nodal_harmonic = foreman_nodal_harmonic.astype(float)
    foreman_nodal_harmonic.index.name = None
    
    return foreman_doodson_harmonic, foreman_nodal_harmonic


@functools.lru_cache() #caching useful since reading file and internal consistency check once is sufficient
def get_foreman_shallowrelations():
    """
    Omzetten van het derde deel van de foremantabel in een pandas DataFrame met shallow water relations.

    Returns
    -------
    foreman_shallowrelations : TYPE
        DESCRIPTION.

    """
    
    foreman_file = os.path.join(os.path.dirname(file_path),'data','data_foreman_shallowrelations.txt')
    foreman_shallowrelations = pd.read_csv(foreman_file, comment='#', names=range(10), index_col=0, sep="\\s+")
    foreman_shallowrelations.index.name = None
    
    #check for internal numdependencies consistency
    for const in foreman_shallowrelations.index:
        foreman_shallow_const = foreman_shallowrelations.loc[const].dropna().values
        num_dependencies = foreman_shallow_const[0]
        list_shallow_facs = foreman_shallow_const[1::2]
        list_shallow_deps = foreman_shallow_const[2::2]
        if not list_shallow_deps.shape == list_shallow_facs.shape == (num_dependencies,):
            raise Exception(f'ERROR: shallow relations not internally consistent:\n{foreman_shallowrelations.loc[const]}')
        
    #check whether all dependencies are available as harmonic components
    foreman_doodson_harmonic, foreman_nodal_harmonic = get_foreman_doodson_nodal_harmonic()
    list_shallowdependencies = foreman_shallowrelations[[3,5,7,9]].melt()['value'].dropna().unique()
    bool_shallowdependencies_isin_harmonics = pd.Series(list_shallowdependencies).isin(foreman_doodson_harmonic.index)
    if not bool_shallowdependencies_isin_harmonics.all():
        raise Exception(f'ERROR: not all required shallow dependency components are available:\n{list_shallowdependencies[~bool_shallowdependencies_isin_harmonics]}')

    #convert to actual equations like schureman (for eval function), #TODO: simplify this
    foreman_shallowrelations_inclplusmult = foreman_shallowrelations.copy()
    foreman_shallowrelations_inclplusmult[['+','*']] = '+','*'
    shallow_eqs_pd_foreman = pd.DataFrame()
    shallow_eqs_pd_foreman['shallow_eq'] = foreman_shallowrelations_inclplusmult[[2,'*',3,'+',4,'*',5,'+',6,'*',7,'+',8,'*',9]].astype(str).apply(''.join, axis=1).str.replace('+nan*nan','',regex=False)
    shallow_eqs_pd_foreman['shallow_const'] = shallow_eqs_pd_foreman.index
    shallow_eqs_pd_foreman.index = 'comp_'+shallow_eqs_pd_foreman.index.str.replace('(','_',regex=False).str.replace(')','_',regex=False)#brackets are temporarily removed in order to evaluate functions (replaced by underscore to distinguish between similar component names like MKS2 and M(KS)2

    return shallow_eqs_pd_foreman, foreman_shallowrelations, list_shallowdependencies


@functools.lru_cache()
def get_foreman_doodson_nodal_all_NOTUSED(lat_deg=51.45): #TODO: maybe use this definition to replace harmonic/shallow separate part in freq/v0 and uf definitions. Tricky because of factors and mainly M2=3.5*M1 (non-int factor)
    foreman_doodson_harmonic, foreman_nodal_harmonic = get_foreman_doodson_nodal_harmonic(lat_deg=lat_deg)
    
    #from hatyan.schureman import get_schureman_shallowrelations
    #shallow_eqs_pd = get_schureman_shallowrelations()
    #shallow_eqs_pd = shallow_eqs_pd.loc[~shallow_eqs_pd['shallow_eq'].str.contains('M1')] #TODO: two M1 dependent components cannot be included now
    #shallow_eqs_pd_str = '\n'.join(f'{key} = {val}' for key, val in shallow_eqs_pd['shallow_eq'].iteritems())
    shallow_eqs_pd_foreman, foreman_shallowrelations, list_shallowdependencies = get_foreman_shallowrelations()
    shallow_eqs_pd_str = '\n'.join(f'{key} = {val}' for key, val in shallow_eqs_pd_foreman['shallow_eq'].iteritems()) 

    #TODO: for harmonics, we can work with schureman shallowrelations, but for nodal we need the more tabular foreman layout since we need to append multiple rows of the nodal table (instead of calculating one row)
    foreman_doodson_all = foreman_doodson_harmonic.T.eval(shallow_eqs_pd_str).T
    foreman_doodson_all.rename(index=shallow_eqs_pd_foreman['shallow_const'],inplace=True)
    
    duplicate_PNP1EDN = pd.Series(index=shallow_eqs_pd_foreman['shallow_const'],dtype=int)
    foreman_nodal_all = foreman_nodal_harmonic.copy()
    foreman_nodal_all['fac'] = 1
    for iC,const in enumerate(foreman_shallowrelations.index):
        foreman_nodal_oneconst = pd.DataFrame()
        foreman_shallow_const = foreman_shallowrelations.loc[const].dropna().values
        list_shallow_facs = foreman_shallow_const[1::2]
        list_shallow_deps = foreman_shallow_const[2::2]
        for harm_factor,harm_const in zip(list_shallow_facs,list_shallow_deps):
            foreman_nodal_depconst = foreman_nodal_harmonic.loc[[harm_const]].copy()
            foreman_nodal_depconst['fac'] = harm_factor
            foreman_nodal_depconst['dep'] = harm_const
            foreman_nodal_oneconst = foreman_nodal_oneconst.append(foreman_nodal_depconst)
        foreman_nodal_oneconst.index = [const]*len(foreman_nodal_oneconst) #overwrite index values with const name
        duplicate_PNP1EDN.loc[const] = foreman_nodal_oneconst.loc[:,['P','N','P1','EDN']].duplicated().sum()
        #foreman_nodal_oneconst = foreman_nodal_oneconst.sort_values('EDN').sort_values('P1').sort_values('N').sort_values('P') #sorting makes duplicates easier to see
        foreman_nodal_all = foreman_nodal_all.append(foreman_nodal_oneconst)
    return foreman_doodson_all, foreman_nodal_all


#################################################
#################### FREQ V0 ####################
#################################################


def get_foreman_v0_freq(const_list, dood_date=pd.DatetimeIndex([dt.datetime(1900,1,1)])):
    """
    Zoekt voor iedere component uit de lijst de v op basis van harmonische doodson getallen en de frequentie rechtstreeks uit de foreman tabel.
    Shallow water componenten worden afgeleid met de relaties beschreven in de foreman tabel.
    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='foreman') #TODO: move check to central location when part of hatyan_settings()?

    #get freq/v0 for harmonic components
    foreman_doodson_harmonic, foreman_nodal_harmonic = get_foreman_doodson_nodal_harmonic()
    #foreman_doodson_all, foreman_nodal_all = get_foreman_doodson_nodal_all()
    #foreman_doodson_harmonic, foreman_nodal_harmonic = foreman_doodson_all, foreman_nodal_all #TODO: now for renaming convenience, but fix names

    doodson_pd = get_doodson_eqvals(dood_date=pd.DatetimeIndex([dt.datetime(1900,1,1)]), mode='freq') #TODO: get freq on multiple dates?
    multiply_variables = doodson_pd.loc[['T','S','H','P','P1'],:]
    t_const_freq_dood = np.dot(foreman_doodson_harmonic.loc[:,['T','S','H','P','P1']],multiply_variables) / (2*np.pi)
    foreman_freqs = pd.DataFrame({'freq':t_const_freq_dood[:,0]},index=foreman_doodson_harmonic.index)
    
    doodson_pd = get_doodson_eqvals(dood_date=dood_date, mode=None)
    multiply_variables = doodson_pd.loc[['T','S','H','P','N','P1'],:]
    v_0i_rad = np.dot(foreman_doodson_harmonic.loc[:,['T','S','H','P','N','P1']],multiply_variables) + 2*np.pi*foreman_doodson_harmonic.loc[:,['EDN']].values
    v_0i_rad_harmonic_pd = pd.DataFrame(v_0i_rad,index=foreman_doodson_harmonic.index)
    
    #derive freq/v0 for shallow water components
    shallow_eqs_pd_foreman, foreman_shallowrelations, list_shallowdependencies = get_foreman_shallowrelations()
    v_0i_rad = pd.DataFrame(np.zeros((len(const_list),len(dood_date))),index=const_list)
    t_const_freq = pd.DataFrame({'freq':np.zeros((len(const_list)))},index=const_list)
    
    #v and freq for harmonic and shallow constituents
    for iC,const in enumerate(const_list):
        if const in v_0i_rad_harmonic_pd.index:
            v_0i_rad.loc[const] = v_0i_rad_harmonic_pd.loc[const]
            t_const_freq.loc[const,'freq'] = foreman_freqs.loc[const,'freq']
        elif const in foreman_shallowrelations.index: #or is not in foreman_harmonic_doodson_all_list
            #raise Exception('this part should not be reached') #TODO: remove this elif part
            v_0i_rad_temp = 0
            t_const_freq_temp = 0
            foreman_shallow_const = foreman_shallowrelations.loc[const].dropna().values
            list_shallow_facs = foreman_shallow_const[1::2]
            list_shallow_deps = foreman_shallow_const[2::2]
            for harm_factor,harm_const in zip(list_shallow_facs,list_shallow_deps):
                v_dependency = v_0i_rad_harmonic_pd.loc[harm_const].values
                freq_dependency = foreman_freqs.loc[harm_const,'freq'] #should be dependent on harmonic doodson numbers (make foreman_freqs_dood_all variable in foreman.py, in freq or harmonic definition)
                v_0i_rad_temp += harm_factor*v_dependency
                t_const_freq_temp += harm_factor*freq_dependency
            v_0i_rad.loc[const,:] = v_0i_rad_temp
            t_const_freq.loc[const,'freq'] = t_const_freq_temp
    
    return v_0i_rad, t_const_freq


#################################################
################# NODALFACTORS ##################
#################################################

def get_foreman_nodalfactors(const_list, dood_date):
    """
    Zoekt voor iedere component uit de lijst de u en f (nodal factors) op basis van satellite doodson getallen uit de foreman tabel.
    Shallow water componenten worden afgeleid met de relaties beschreven in de foreman tabel.
    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='foreman') #TODO: move check to central location when part of hatyan_settings()?
    
    doodson_pd = get_doodson_eqvals(dood_date)
    foreman_doodson_harmonic, foreman_nodal_harmonic = get_foreman_doodson_nodal_harmonic()
    #foreman_doodson_all, foreman_nodal_all = get_foreman_doodson_nodal_all()
    #foreman_doodson_harmonic, foreman_nodal_harmonic = foreman_doodson_all, foreman_nodal_all #TODO: now for renaming convenience, but fix names
    shallow_eqs_pd_foreman, foreman_shallowrelations, list_shallowdependencies = get_foreman_shallowrelations()
    
    #extent const_list with missing but required harmonic components
    bool_shallowrequired = pd.Series(foreman_shallowrelations.index).isin(pd.Series(const_list))
    bool_dependencyalready = pd.Series(list_shallowdependencies).isin(pd.Series(const_list))
    if bool_shallowrequired.any() and not bool_dependencyalready.all(): #if any shallow constituent is requested and if not all dependent constituents are in const_list
        const_list_extra = list_shallowdependencies[~bool_dependencyalready]
        const_list_inclshallow = const_list + const_list_extra.tolist() # const_list + not yet available shallowdependencies
    else:
        const_list_inclshallow = const_list
    
    #allocate dataframes including potential exta constituents (are removed before return)
    #TODO: creating empty dataframe with {} and columns/index takes way more time than below, check if this is done somewhere in the code
    nodal_shape = (len(const_list_inclshallow),len(dood_date))
    f_i_FOR_inclshallow = pd.DataFrame(np.ones(nodal_shape), index=const_list_inclshallow)
    u_i_rad_FOR_inclshallow = pd.DataFrame(np.zeros(nodal_shape), index=const_list_inclshallow)
        
    #f and u for harmonic constituents
    for const in const_list_inclshallow:
        if const in foreman_nodal_harmonic.index.unique(): # if harmonic constituent has nodal factors
            foreman_harmonic_nodal_const = foreman_nodal_harmonic.loc[[const]]
            fore_delta_jk_rad_all = np.dot(foreman_harmonic_nodal_const.loc[:,['P','N','P1']],doodson_pd.loc[['P','N','P1'],:])
            fore_alpha_jk_all = foreman_harmonic_nodal_const.loc[:,['EDN']].values * 2*np.pi #phase correction satellite. 0.5=90deg voor M2 en N2, 0=0deg voor S2
            fore_r_jk_all = foreman_harmonic_nodal_const.loc[:,['factor']].values #amplitude ratio for satellite. 0.0373 voor M2 en N2, 0.0022 voor S2
            fore_fj_left_all = 1 * fore_r_jk_all * np.cos(fore_delta_jk_rad_all + fore_alpha_jk_all) #should be sum for n sattelites
            fore_fj_right_all = 1 * fore_r_jk_all * np.sin(fore_delta_jk_rad_all + fore_alpha_jk_all) #should be sum for n sattelites
            fore_fj_left = fore_fj_left_all.sum(axis=0)
            fore_fj_right = fore_fj_right_all.sum(axis=0)
            f_i_FOR_inclshallow.loc[const,:] = ( (1+fore_fj_left)**2 + (fore_fj_right)**2)**(1/2.) 
            u_i_rad_FOR_inclshallow.loc[const,:] = -np.arctan2(fore_fj_right,1+fore_fj_left) #TODO: added minus to get sign comparable to hatyan
        elif const in foreman_doodson_harmonic.index: #if harmonic constituent has no nodal factors, default values 1 and 0 are applicable
            continue
    
    #f and u for shallow constituents
    for const in const_list_inclshallow:#TODO: remove this part, currently the foreman_nodal_all is constructed based on + and -, but for f it should be all plusses
        if const in foreman_shallowrelations.index: # component has satellites based on shallow water relations
            f_i_FOR_temp = 1.0
            u_i_rad_FOR_temp = 0.0
            foreman_shallow_const = foreman_shallowrelations.loc[const].dropna().values
            list_shallow_facs = foreman_shallow_const[1::2]
            list_shallow_deps = foreman_shallow_const[2::2]
            for harm_factor,harm_const in zip(list_shallow_facs,list_shallow_deps):
                f_i_FOR_temp *= f_i_FOR_inclshallow.loc[harm_const,:].values ** abs(harm_factor)
                u_i_rad_FOR_temp += u_i_rad_FOR_inclshallow.loc[harm_const,:].values * harm_factor
            f_i_FOR_inclshallow.loc[const,:] = f_i_FOR_temp
            u_i_rad_FOR_inclshallow.loc[const,:] = u_i_rad_FOR_temp
    
    #drop extra added constituents
    f_i_FOR = f_i_FOR_inclshallow.loc[const_list,:]
    u_i_rad_FOR = u_i_rad_FOR_inclshallow.loc[const_list,:]
    return f_i_FOR, u_i_rad_FOR





# -*- coding: utf-8 -*-
"""
Created on Thu Jan 21 11:17:44 2021

@author: laan_st

"""


import pytest
import numpy as np
import pandas as pd
import datetime as dt
import hatyan


@pytest.mark.unittest
def test_astrog_deprecated_tzone_argument():
    func_list = [
        hatyan.astrog.astrog_culminations,
        hatyan.astrog.astrog_phases,
        hatyan.astrog.astrog_sunriseset,
        hatyan.astrog.astrog_moonriseset,
        hatyan.astrog.astrog_anomalies,
        hatyan.astrog.astrog_seasons,
    ]
    for func in func_list:
        with pytest.raises(DeprecationWarning) as e:
            func(tFirst="2000-01-01", tLast="2001-01-01", tzone="UTC+01:00")
        assert "Argument 'tzone' has been deprecated for" in str(e.value)


@pytest.mark.unittest
def test_astrog_convert_str2datetime_str_tzdifferent():
    with pytest.raises(AssertionError):
        hatyan.astrog.convert_str2datetime("1887-01-01 00:00 +00:00",
                                           "2022-12-31 00:00 +01:00")


@pytest.mark.unittest
def test_astrog_convert_str2datetime_enddate_toosmall():
    with pytest.raises(ValueError) as e:
        hatyan.astrog.convert_str2datetime("2022-01-01 00:00 +00:00",
                                           "2010-12-31 00:00 +00:00")
    assert "start_date 2022-01-01 00:00:00 is larger than end_date 2010-12-31 00:00:00" in str(e.value)


@pytest.mark.unittest
def test_astrog_convert_str2datetime_str_naive():
    tstart, tstop, tzone = hatyan.astrog.convert_str2datetime("1887-01-01","2022-12-31")
    assert tstart == pd.Timestamp("1887-01-01")
    assert tstart.tz is None
    assert tstop == pd.Timestamp("2022-12-31")
    assert tstop.tz is None
    assert tzone == "UTC"


@pytest.mark.unittest
def test_astrog_convert_str2datetime_pd_naive():
    tstart, tstop, tzone = hatyan.astrog.convert_str2datetime(pd.Timestamp("1887-01-01"),
                                                              pd.Timestamp("2022-12-31"))
    assert tstart == pd.Timestamp("1887-01-01")
    assert tstart.tz is None
    assert tstop == pd.Timestamp("2022-12-31")
    assert tstop.tz is None
    assert tzone == "UTC"


@pytest.mark.unittest
def test_astrog_convert_str2datetime_str_met():
    tstart, tstop, tzone = hatyan.astrog.convert_str2datetime("1887-01-01 00:00 +01:00",
                                                              "2022-12-31 00:00 +01:00")
    assert tstart == pd.Timestamp("1886-12-31 23:00:00")
    assert tstart.tz is None
    assert tstop == pd.Timestamp("2022-12-30 23:00:00")
    assert tstop.tz is None
    assert tzone == dt.timezone(dt.timedelta(seconds=3600))


@pytest.mark.unittest
def test_astrog_convert_str2datetime_pd_met():
    tstart, tstop, tzone = hatyan.astrog.convert_str2datetime(pd.Timestamp("1887-01-01 00:00 +01:00"),
                                                              pd.Timestamp("2022-12-31 00:00 +01:00"))
    assert tstart == pd.Timestamp("1886-12-31 23:00:00")
    assert tstart.tz is None
    assert tstop == pd.Timestamp("2022-12-30 23:00:00")
    assert tstop.tz is None
    assert tzone == dt.timezone(dt.timedelta(seconds=3600))


@pytest.mark.unittest
def test_astrog_convert_str2datetime_pd_utc():
    tstart, tstop, tzone = hatyan.astrog.convert_str2datetime(pd.Timestamp("1887-01-01 00:00 +00:00"),
                                                              pd.Timestamp("2022-12-31 00:00 +00:00"))
    assert tstart == pd.Timestamp("1887-01-01")
    assert tstart.tz is None
    assert tstop == pd.Timestamp("2022-12-31")
    assert tstop.tz is None
    assert tzone == dt.timezone.utc
    

@pytest.mark.systemtest
def test_astrog_dT():
    # 1. Input
    timeInput = pd.date_range(start=dt.datetime(1970,12,31,12,31),end=dt.datetime(2020,12,31,12,31),freq='YS')
    
    # 2. Expectations
    dTLastExp = np.array([39.468, 40.144, 40.82 , 41.496, 42.172, 42.848, 43.524, 44.2  ,
           44.876, 45.552, 46.228, 46.904, 47.58 , 48.256, 48.932, 49.608,
           50.284, 50.96 , 51.636, 52.312, 52.988, 53.664, 54.34 , 55.016,
           55.692, 56.368, 57.044, 57.72 , 58.396, 59.072, 59.748, 60.424,
           61.1  , 61.776, 62.452, 63.128, 63.804, 64.48 , 65.156, 65.832,
           66.508, 67.184, 67.86 , 68.536, 69.212, 69.888, 70.564, 71.24 ,
           71.916, 72.592])
    dTExctExp = np.array([42.184, 42.184, 44.184, 45.184, 46.184, 47.184, 48.184, 49.184,
           50.184, 51.184, 51.184, 52.184, 53.184, 54.184, 54.184, 55.184,
           55.184, 56.184, 56.184, 57.184, 58.184, 58.184, 59.184, 60.184,
           61.184, 62.184, 62.184, 63.184, 64.184, 64.184, 64.184, 64.184,
           64.184, 64.184, 64.184, 65.184, 65.184, 65.184, 66.184, 66.184,
           66.184, 66.184, 67.184, 67.184, 67.184, 68.184, 69.184, 69.184,
           69.184, 69.184])
    
    # 3. Run test
    dTLastOut = hatyan.astrog.dT(timeInput,dT_fortran=True)
    dTExctOut = hatyan.astrog.dT(timeInput,dT_fortran=False)
    
    # 4. Vefication
    assert type(dTLastOut) == np.ndarray
    assert type(dTExctOut) == np.ndarray
    assert np.allclose(dTLastExp, dTLastOut)
    assert np.allclose(dTExctExp, dTExctOut)


@pytest.mark.systemtest
def test_astrog_astrab():
    
    # 1. Input
    timeInput = dt.datetime(2008,1,1,12,31)
    
    # 2. Expectations
    parExpect = {'EHMOON': np.array([448.08691901]),
                 'DECMOO': np.array([-13.02683311]),
                 'PARLAX': np.array([3262.092635]),
                 'DPAXDT': np.array([-18.05431966]),
                 'ALTMOO': np.array([-13.09124181]),
                 'ELONG':  np.array([283.08899401]),
                 'ALTSUN': np.array([14.08083711]),
                 'LONSUN': np.array([280.45809545]),
                 'EQELON': np.array([278.83322595]),
                 'DECSUN': np.array([-23.02828019]),
                 'DISSUN': np.array([0.98330485]),
                 'EHARI':  np.array([5.0317035]),
                 'RASUN':  np.array([281.37522918]),
                 'EHSUN':  np.array([6.92014497]),
                 'LONMOO': np.array([203.54708946]),
                 'LATMOO': np.array([-4.18226297]),
                 'RAMOON': np.array([200.20845513]),
                 'ANM':    np.array([151.15606465])}
    
    # 3. Run test
    parOutput = hatyan.astrog.astrab(timeInput,hatyan.astrog.dT(timeInput,dT_fortran=True))
    
    # 4. Vefication
    assert type(parOutput) == dict
    assert len(parOutput)  == 18
    for parName in parExpect:
        assert len(parOutput[parName]) == 1
        assert abs(parExpect[parName][0]-parOutput[parName][0]) < 10E-9
    
    
@pytest.mark.systemtest
def test_astrog_astrac():
    
    # 1. Input
    timeInput  = dt.datetime(2008,1,1,6,31)
    
    # 2. Expectations
    timeExpect = [pd.DatetimeIndex(['2007-12-31 18:08:06.634848001']),
                  pd.DatetimeIndex(['2008-01-01 06:28:32.856047999']),
                  pd.DatetimeIndex(['2007-12-17 10:18:26.216188801']),
                  pd.DatetimeIndex(['2007-12-24 01:16:42.818976000']),
                  pd.DatetimeIndex(['2007-12-31 07:51:54.693312000']),
                  pd.DatetimeIndex(['2008-01-08 11:38:04.174895999']),
                  pd.DatetimeIndex(['2008-01-01 00:54:56.645894400']),
                  pd.DatetimeIndex(['2008-01-01 11:07:33.697843200']),
                  pd.DatetimeIndex(['2008-01-01 07:48:34.139385600']),
                  pd.DatetimeIndex(['2008-01-01 07:48:34.139471999']),
                  pd.DatetimeIndex(['2008-03-20 05:49:10.054166400']),
                  pd.DatetimeIndex(['2007-06-21 18:07:03.793017600']),
                  pd.DatetimeIndex(['2007-09-23 09:52:02.142432000']),
                  pd.DatetimeIndex(['2007-12-22 06:08:46.055884800']),
                  pd.DatetimeIndex(['2008-01-03 08:22:45.903254400']),
                  pd.DatetimeIndex(['2008-01-03 08:22:45.696239999'])]
     
    # 3. Run test
    timeOutput = []
    for iMode in range(1,17):
        timeOutput.append(hatyan.astrog.astrac(timeInput,dT_fortran=False,mode=np.array(iMode)))
    
    # 4. Vefication
    for iMode in range(1,17):
        print(iMode)
        assert type(timeOutput[iMode-1]) == pd.DatetimeIndex
        assert abs(timeExpect[iMode-1]-timeOutput[iMode-1]).total_seconds() < 10E-5


@pytest.mark.systemtest
def test_astrog_leapsecondslist():
    leap_seconds_pd, expirydate = hatyan.astrog.get_leapsecondslist_fromurlorfile()


@pytest.mark.systemtest
def test_astrog_culminations():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # moon culminations
    culminations_python = hatyan.astrog_culminations(tFirst=start_date, tLast=end_date)
    
    datetimes = culminations_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-01-01 04:44:14.732116514+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-31 13:16:46.464269856+0000')

    subset = culminations_python.iloc[:10]
    assert subset["type"].tolist() == [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    expected_type_str = [
        'lowerculmination', 'upperculmination',
        'lowerculmination', 'upperculmination',
        'lowerculmination', 'upperculmination',
        'lowerculmination', 'upperculmination',
        'lowerculmination', 'upperculmination',
        ]
    assert subset["type_str"].tolist() == expected_type_str
    expected_parallax = np.array([
        0.90436235, 0.90347198, 0.90333631, 0.90397446, 0.90539423,
           0.90759186, 0.9105516 , 0.91424508, 0.91863045, 0.92365127])
    assert np.allclose(subset["parallax"].values, expected_parallax)
    expected_declination = np.array([
        -9.12903352, -6.86053844, -4.53030521, -2.15683332,  0.24204067,
        2.64853148,  5.04424777,  7.40950843,  9.72265897, 11.95942792])
    assert np.allclose(subset["declination"].values, expected_declination)


@pytest.mark.systemtest
def test_astrog_culminations_met():
    start_date = "2020-01-01 00:00:00 +01:00"
    end_date = "2021-01-01 00:00:00 +01:00"
    # moon culminations
    culminations_python = hatyan.astrog_culminations(tFirst=start_date, tLast=end_date)
    
    datetimes = culminations_python.index
    assert datetimes.tz == dt.timezone(dt.timedelta(seconds=3600))
    assert datetimes[0] == pd.Timestamp('2020-01-01 05:44:14.732115815+0100')
    assert datetimes[-1] == pd.Timestamp('2020-12-31 14:16:46.464269693+0100')


@pytest.mark.systemtest
def test_astrog_phases():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # lunar phases
    phases_python = hatyan.astrog_phases(tFirst=start_date, tLast=end_date)
    
    datetimes = phases_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-01-03 04:45:20+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-30 03:28:07+0000')
    
    expected_type = [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 
                     1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 
                     1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 
                     1, 2]
    assert phases_python["type"].tolist() == expected_type
    expected_type_str = ['FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM', 'LQ', 'NM', 'FQ', 'FM', 'LQ', 'NM', 
                         'FQ', 'FM']
    assert phases_python["type_str"].tolist() == expected_type_str


@pytest.mark.systemtest
def test_astrog_moonriseset():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # moonrise and -set
    moonriseset_python = hatyan.astrog_moonriseset(tFirst=start_date, tLast=end_date)
    moonriseset_python_perday = hatyan.convert2perday(moonriseset_python)
    
    datetimes = moonriseset_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-01-01 11:15:43+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-31 16:46:52+0000')
    
    datetimes_perday = pd.to_datetime(moonriseset_python_perday.index)
    assert datetimes_perday.tz is None
    assert datetimes_perday[0] == pd.Timestamp('2020-01-01 00:00:00')
    assert datetimes_perday[-1] == pd.Timestamp('2020-12-31 00:00:00')
    
    subset = moonriseset_python.iloc[:10]
    assert subset["type"].tolist() == [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    expected_type_str = [
        'moonrise', 'moonset', 'moonrise', 'moonset',
        'moonrise', 'moonset', 'moonrise', 'moonset',
        'moonrise', 'moonset']
    assert subset["type_str"].tolist() == expected_type_str
    
    subset_perday = moonriseset_python_perday.iloc[:10]
    assert subset_perday["moonrise"].isnull().sum() == 0
    assert subset_perday["moonset"].isnull().sum() == 1


@pytest.mark.systemtest
def test_astrog_moonriseset_posinf():
    """
    This part of code resulted in timesteps being dropped in newer pandas versions.
    This was due to 0 RATES in astract, resulting in inf values in 'addtime'.
    This is solved by supplying posinf=0 to np.nan_to_num().
    This test therefore checks if the length of the resulting dataframe is correct.
    """
    
    # script settings
    tstart = dt.datetime(2003,1,1)
    tstop = dt.datetime(2003,6,1)
    
    # moonrise and -set
    moonriseset_python = hatyan.astrog_moonriseset(tFirst=tstart, tLast=tstop)
    
    assert len(moonriseset_python) == 292


@pytest.mark.systemtest
def test_astrog_sunriseset():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # sunrise and -set
    sunriseset_python = hatyan.astrog_sunriseset(tFirst=start_date, tLast=end_date)
    sunriseset_python_perday = hatyan.convert2perday(sunriseset_python)
    
    datetimes = sunriseset_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-01-01 07:47:25+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-31 15:36:02+0000')
    
    datetimes_perday = pd.to_datetime(sunriseset_python_perday.index)
    assert datetimes_perday.tz is None
    assert datetimes_perday[0] == pd.Timestamp('2020-01-01 00:00:00')
    assert datetimes_perday[-1] == pd.Timestamp('2020-12-31 00:00:00')

    subset = sunriseset_python.iloc[:10]
    assert subset["type"].tolist() == [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    expected_type_str = [
        'sunrise', 'sunset', 'sunrise', 'sunset', 
        'sunrise', 'sunset', 'sunrise', 'sunset', 
        'sunrise', 'sunset']
    assert subset["type_str"].tolist() == expected_type_str
    
    subset_perday = sunriseset_python_perday.iloc[:10]
    assert subset_perday["sunrise"].isnull().sum() == 0
    assert subset_perday["sunset"].isnull().sum() == 0
    

@pytest.mark.systemtest
def test_astrog_anomalies():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # lunar anomalies
    anomalies_python = hatyan.astrog_anomalies(tFirst=start_date, tLast=end_date)
    
    datetimes = anomalies_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-01-02 01:25:44+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-24 16:48:18+0000')
    
    expected_type = [2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 
                     1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    expected_type_str = ['apogeum', 'perigeum', 'apogeum', 'perigeum', 'apogeum', 'perigeum', 
                         'apogeum', 'perigeum', 'apogeum', 'perigeum', 'apogeum', 'perigeum', 
                         'apogeum', 'perigeum', 'apogeum', 'perigeum', 'apogeum', 'perigeum', 
                         'apogeum', 'perigeum', 'apogeum', 'perigeum', 'apogeum', 'perigeum', 
                         'apogeum', 'perigeum', 'apogeum']
    assert anomalies_python['type'].tolist() == expected_type
    assert anomalies_python['type_str'].tolist() == expected_type_str


@pytest.mark.systemtest
def test_astrog_seasons():
    start_date = "2020-01-01"
    end_date = "2021-01-01"
    # astronomical seasons
    seasons_python = hatyan.astrog_seasons(tFirst=start_date, tLast=end_date)
    
    datetimes = seasons_python.index
    assert datetimes.tz == dt.timezone.utc
    assert datetimes[0] == pd.Timestamp('2020-03-20 03:49:46+0000')
    assert datetimes[-1] == pd.Timestamp('2020-12-21 10:02:39+0000')
    
    assert seasons_python['type'].tolist() == [1, 2, 3, 4]
    assert seasons_python['type_str'].tolist() == ['spring', 'summer', 'autumn', 'winter']


@pytest.mark.unittest
def test_plot_astrog_diff():
    start_date_utc = pd.Timestamp(2000, 1, 1, tz="UTC")
    end_date_utc = pd.Timestamp(2000, 4, 1, tz="UTC")
    
    culminations_python = hatyan.astrog_culminations(tFirst=start_date_utc, tLast=end_date_utc)
    culminations_python_naive = culminations_python.tz_localize(None).reset_index()
    
    hatyan.plot_astrog_diff(culminations_python, culminations_python_naive, typeLab=['lower','upper'], timeBand=[-.18,.18])



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 25 09:25:17 2024

@author: veenstra
"""

import os
import pytest
from click.testing import CliRunner
from hatyan import cli
import shutil
import glob

dir_tests = os.path.dirname(__file__) #F9 doesnt work, only F5 (F5 also only method to reload external definition scripts)

@pytest.mark.unittest
def test_command_line_interface():
    """Test the CLI."""
    filename = os.path.join(dir_tests,'examples','minimal_example.py')
    assert os.path.exists(filename)
    
    runner = CliRunner()
    
    result = runner.invoke(cli.cli)
    assert result.exit_code == 2
    assert 'Missing argument' in result.output
    
    help_result = runner.invoke(cli.cli, ['--help'])
    assert help_result.exit_code == 0
    assert 'Show this message and exit.' in help_result.output
    
    dir_output = os.path.abspath(os.path.basename(filename).replace(".py",""))
    if os.path.exists(dir_output):
        shutil.rmtree(dir_output)
    filename_command = f'{filename} --redirect-stdout'
    filename_result = runner.invoke(cli.cli, filename_command.split())
    assert filename_result.exit_code == 0
    
    # assert file presence
    file_stdout = os.path.join(dir_output, "STDOUT.txt")
    file_script = os.path.join(dir_output, os.path.basename(filename))
    dia_list = glob.glob(os.path.join(dir_output, "*.dia"))
    png_list = glob.glob(os.path.join(dir_output, "*.png"))
    
    assert os.path.exists(dir_output)
    assert os.path.exists(file_stdout)
    assert os.path.exists(file_script)
    assert len(dia_list) > 0
    assert len(png_list) > 0
    
    # cleanup output
    shutil.rmtree(dir_output)



# - * - coding: utf - 8 - * - 
"""
schureman.py contains definitions with data for the schureman constituents.
"""

import os
import pandas as pd
import functools #to Memoize v0uf table (https://en.wikipedia.org/wiki/Memoization)
import numpy as np
import datetime as dt
import logging
import warnings

file_path = os.path.realpath(__file__)
logger = logging.getLogger(__name__)


@functools.lru_cache()
def get_schureman_shallowrelations():
    file_schureman_shallowrelations = os.path.join(os.path.dirname(file_path),'data','data_schureman_shallowrelations.csv')
    shallow_eqs_pd = pd.read_csv(file_schureman_shallowrelations,comment='#',skipinitialspace=True,index_col=0,names=['shallow_eq'])
    shallow_eqs_pd['shallow_eq'] = shallow_eqs_pd['shallow_eq'].str.strip() #remove spaces after
    shallow_eqs_pd['shallow_const'] = shallow_eqs_pd.index
    
    shallow_eqs_pd.index = 'comp_'+shallow_eqs_pd.index.str.replace('(','_',regex=False).str.replace(')','_',regex=False)#brackets are temporarily removed in order to evaluate functions (replaced by underscore to distinguish between similar component names like MKS2 and M(KS)2
    return shallow_eqs_pd


@functools.lru_cache()
def get_schureman_table():
    """
    Calculate all schureman constituents

    Returns
    -------
    v0uf_all : TYPE
        DESCRIPTION.
    
    """
    
    index_v0 = ['T','S','H','P','N','P1','EDN']
    index_u = ['DKSI','DNU','DQ','DQU','DR','DUK1','DUK2']
    index_f = ['DND73','DND74','DND75','DND76','DND77','DND78','DND79','DFM1','DFK1','DFL2','DFK2','DFM1C']

    file_schureman_harmonic = os.path.join(os.path.dirname(file_path),'data','data_schureman_harmonic.csv')
    v0uf_baseT = pd.read_csv(file_schureman_harmonic,comment='#',skipinitialspace=True,index_col='component')
    v0uf_base = v0uf_baseT.T
    
    shallow_eqs_pd = get_schureman_shallowrelations()
    shallow_eqs_pd_str = '\n'.join(f'{key} = {val}' for key, val in shallow_eqs_pd['shallow_eq'].items()) 

    #calculate shallow water components and rename back to original component name
    v0uf_base_forv0u = v0uf_base.loc[index_v0+index_u,:].astype(int)
    v0uf_base_forf = v0uf_base.loc[index_f,:].astype(float)
    with warnings.catch_warnings():
        # suppress PerformanceWarning: https://github.com/Deltares/hatyan/issues/272
        warnings.simplefilter("ignore", category=pd.errors.PerformanceWarning)
        v0uf_base_forv0u = v0uf_base_forv0u.eval(shallow_eqs_pd_str, inplace=False).copy()
        # for f only multiplication is applied, never division
        v0uf_base_forf = v0uf_base_forf.eval(shallow_eqs_pd_str.replace('-','+'), inplace=False).copy()
    v0uf_all = pd.concat([v0uf_base_forv0u,v0uf_base_forf])
    v0uf_all.rename(columns=shallow_eqs_pd['shallow_const'],inplace=True)
    v0uf_allT = v0uf_all.T
    
    return v0uf_allT


def get_schureman_freqs(const_list, dood_date=pd.DatetimeIndex([dt.datetime(1900,1,1)]), return_allraw=False):
    """
    Returns the frequencies of the requested list of constituents. Source: beromg.f

    Parameters
    ----------
    const_list : list or pandas.Series
        contains the tidal constituent names.

    Raises
    ------
    Exception
        DESCRIPTION.

    Returns
    -------
    t_const_freq : TYPE
        DESCRIPTION.

    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='schureman') #TODO: move check to central location when part of hatyan_settings()?
    
    doodson_pd = get_doodson_eqvals(dood_date=dood_date, mode='freq') #N is not used here
    multiply_variables = doodson_pd.loc[['T','S','H','P','P1'],:]
    
    v0uf_allT = get_schureman_table()
    v0uf_sel = v0uf_allT.loc[const_list]
    v0uf_sel_freq = v0uf_sel[['T','S','H','P','P1']]
    
    DOMEGA_speed = np.dot(v0uf_sel_freq,multiply_variables)
    if return_allraw: #return array of speeds for each component/timestep
        return DOMEGA_speed
    
    #return dataframe of freq/angvelo/period/angfreq for all components and first timestep
    t_const_speed = DOMEGA_speed[:,0]
    t_const_freq = t_const_speed/(2*np.pi) #aantal rotaties per uur, freq
    np.seterr(divide='ignore') #suppress divide by 0 warning
    t_const_perds = 1/t_const_freq #period [hr]
    t_const_angfreqs = 360/t_const_perds #angfreq [deg/hr]
    freq_pd = pd.DataFrame({'freq':t_const_freq, 'angvelo [rad/hr]':t_const_speed, 'period [hr]':t_const_perds, 'angfreq [deg/hr]':t_const_angfreqs}, index=const_list)
    return freq_pd


def get_schureman_v0(const_list, dood_date):
    """
    Returns the v-values of the requested list of constituents for the requested date(s)
    
    Parameters
    ----------
    const_list : list or pandas.Series
        contains the tidal constituent names.
    dood_date : TYPE
        DESCRIPTION.

    Returns
    -------
    v_0i_rad : TYPE
        DESCRIPTION.

    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='schureman') #TODO: move check to central location when part of hatyan_settings()?

    doodson_pd = get_doodson_eqvals(dood_date=dood_date) #N is not used here
    multiply_variables = doodson_pd.loc[['T','S','H','P','P1'],:]
    
    v0uf_allT = get_schureman_table()
    v0uf_sel = v0uf_allT.loc[const_list]
    DV0 = np.dot(v0uf_sel.loc[:,['T','S','H','P','P1']],multiply_variables) + np.deg2rad(v0uf_sel.loc[:,['EDN']]).values
    DV0_pd = pd.DataFrame(DV0,index=const_list)
    
    return DV0_pd


def get_schureman_constants(dood_date):
    """
    get_schureman_constants

    Parameters
    ----------
    dood_date : TYPE
        DESCRIPTION.

    Returns
    -------
    DOMEGA : TYPE
        DESCRIPTION.
    DIKL : TYPE
        DESCRIPTION.
    DC1681 : TYPE
        DESCRIPTION.
    DC5023 : TYPE
        DESCRIPTION.
    DC0365 : TYPE
        DESCRIPTION.

    """
    
    #TODO: robust_timedelta_sec might also be necesary in other definitions, but is not there yet. Align? (also check if newer pandas version do not have this problem anymore)
    from hatyan.hatyan_core import robust_timedelta_sec # local import since otherwise cross-dependency
    
    #bercon.f: HET BEREKENEN VAN DE 'CONSTANTEN' .0365, .1681 EN .5023, DIE GEBRUIKT WORDEN BIJ DE BEREKENING VAN DE U- EN F-FACTOREN VAN DE GETIJCOMPONENTEN K1 EN K2
    #327932: ratio of mass of sun to combined mass of earth and moon
    #81.53: ratio of mass of earth to mass of moon
    DAGC = 0.01657                            #A/C >> ZIE BLZ. 162 VAN SCHUREMAN, mean lunar parallax in respect to mean radius [rad]
    DAGC1 = 0.00004261                        #A/C1 >> ZIE BLZ. 162 VAN SCHUREMAN, mean solar parallax in respect to mean radius [rad]
    DE = 0.054900489                          #E >> ZIE BLZ. 162 VAN SCHUREMAN, eccentricity of moons orbit
    DMGE = 1./81.53                           #M/E >> ZIE BLZ. 162 VAN SCHUREMAN, mass of moon /mass of earth
    DSGE = 82.53/81.53*327932                 #S/E >> ZIE BLZ. 162 VAN SCHUREMAN, mass of sun / mass of earth
    
    DU = DMGE*DAGC**3                         #U >> ZIE BLZ. 162 VAN SCHUREMAN, basic factor
    DU1 = DSGE*DAGC1**3                       #U1 >> ZIE BLZ. 162 VAN SCHUREMAN, solar coefficient
    DSACCE = DU1/DU                           #S' >> ZIE BLZ. 162 VAN SCHUREMAN, solar factor
    DOMEGA = np.deg2rad(23+27./60.+8.26/3600) #OMEGA >> ZIE BLZ. 162 VAN SCHUREMAN, Obliquity of the Ecliptic, epoch 1 Jan 1900, 23.45229 graden [rad]
    DIKL = np.deg2rad(5+8./60.+43.3546/3600)  #i >> ZIE BLZ. 162 VAN SCHUREMAN, Inclination of moons orbit to plane of ecliptic, epoch 1 Jan 1900 (?), 5.1453762 graden [rad]
    
    DC5023 = 0.5+0.75*DE*DE

    #time dependent
    dood_tstart_sec = robust_timedelta_sec(dood_date)    
    dood_Tj = (dood_tstart_sec/3600+12)/(24*36525)
    DE1 = 0.01675104-0.0000418*dood_Tj        #E1 >> ZIE BLZ. 162 VAN SCHUREMAN, eccentricity of earths orbit, epoch 1 Jan 1900
    DCOFSI = (0.5+0.75*DE1**2)*DSACCE         #COEFFICIENT VAN DE SINUSTERMEN IN (217) EN (219) OP BLZ. 45 VAN SCHUREMAN

    DC0365 = DCOFSI*np.sin(DOMEGA)**2
    DC1681 = DCOFSI*np.sin(2*DOMEGA)
    
    return DOMEGA, DIKL, DC1681, DC5023, DC0365


def get_schureman_u(const_list, dood_date):
    """
    Returns the u-values of the requested list of constituents for the requested date(s)

    Parameters
    ----------
    const_list : list or pandas.Series
        contains the tidal constituent names.
    dood_date : TYPE
        DESCRIPTION.

    Returns
    -------
    u_i_rad_HAT : TYPE
        DESCRIPTION.

    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='schureman') #TODO: move check to central location when part of hatyan_settings()?

    doodson_pd = get_doodson_eqvals(dood_date=dood_date)
    N_rad = doodson_pd.loc['N',:].values
    P_rad = doodson_pd.loc['P',:].values
    
    #list of dependencies for U (only P and N are used here)
    DOMEGA, DIKL, DC1681, DC5023, DC0365 = get_schureman_constants(dood_date)
    DHOMI = (DOMEGA-DIKL)*0.5
    DHOPI = (DOMEGA+DIKL)*0.5
    
    DTHN = np.tan(N_rad*0.5)
    DATC = np.arctan2(np.cos(DHOMI)*DTHN,np.cos(DHOPI))
    DATS = np.arctan2(np.sin(DHOMI)*DTHN,np.sin(DHOPI))
    DIH = np.arccos(np.cos(DIKL)*np.cos(DOMEGA)-np.sin(DIKL)*np.sin(DOMEGA)*np.cos(N_rad))
    DKSI = (N_rad-DATC-DATS)
    DNU = DATC-DATS
    DPMKSI = P_rad-DKSI
    DC2PMK = np.cos(DPMKSI+DPMKSI)
    DS2PMK = np.sin(DPMKSI+DPMKSI)
    DCIH = np.cos(DIH)
    DSIH = np.sin(DIH)
    DCHIH = np.cos(DIH*0.5)
    DCTHIH = 1./np.tan(DIH*0.5)
    DS2IH = np.sin(DIH+DIH)
    DQU = np.arctan2(DS2PMK,(3*DCIH/(DCHIH*DCHIH)+DC2PMK))
    DQ = np.arctan2((5*DCIH-1)*np.tan(DPMKSI),(7*DCIH+1))
    DNUACC = np.arctan2(DS2IH*np.sin(DNU),(DS2IH*np.cos(DNU)+DC1681/DC5023))
    DR = np.arctan2(DS2PMK,(DCTHIH*DCTHIH/6.-DC2PMK))
    D2NU2A = np.arctan2(DSIH**2*np.sin(2*DNU),(DSIH**2*np.cos(2*DNU)+DC0365/DC5023))
    DUK1   = -DNUACC 
    DUK2   = -D2NU2A
    multiply_variables = np.stack([DKSI, DNU, DQ, DQU, DR, DUK1, DUK2])
    
    v0uf_allT = get_schureman_table()
    v0uf_sel = v0uf_allT.loc[const_list]
    v0uf_sel_u = v0uf_sel[['DKSI','DNU', 'DQ', 'DQU', 'DR', 'DUK1', 'DUK2']]

    DU = np.dot(v0uf_sel_u.values,multiply_variables)
    DU = np.remainder(DU+np.pi, 2*np.pi) - np.pi

    #create dataframe
    DU_pd = pd.DataFrame(DU, index=const_list)
    
    return DU_pd


def get_schureman_f(const_list, dood_date, xfac):
    """
    Returns the f-values of the requested list of constituents for the requested date(s)

    Parameters
    ----------
    const_list : list or pandas.Series
        contains the tidal constituent names.
    dood_date : TYPE
        DESCRIPTION.

    Returns
    -------
    f_i_HAT : TYPE
        DESCRIPTION.

    """
    
    from hatyan.hatyan_core import get_doodson_eqvals, check_requestedconsts # local import since otherwise cross-dependency
    
    check_requestedconsts(tuple(const_list),source='schureman') #TODO: move check to central location when part of hatyan_settings()?

    doodson_pd = get_doodson_eqvals(dood_date=dood_date)
    N_rad = doodson_pd.loc['N',:].values
    P_rad = doodson_pd.loc['P',:].values
    
    #list of dependencies F (only P and N are used here)
    DOMEGA, DIKL, DC1681, DC5023, DC0365 = get_schureman_constants(dood_date)
    DHOMI = (DOMEGA-DIKL)*0.5
    DHOPI = (DOMEGA+DIKL)*0.5
    DSOMEG = np.sin(DOMEGA)
    DSIKL = np.sin(DIKL)
    DTHN = np.tan(N_rad*0.5)
    DATC = np.arctan2(np.cos(DHOMI)*DTHN,np.cos(DHOPI))
    DATS = np.arctan2(np.sin(DHOMI)*DTHN,np.sin(DHOPI))
    DIH = np.arccos(np.cos(DIKL)*np.cos(DOMEGA)-DSIKL*DSOMEG*np.cos(N_rad))
    DKSI = (N_rad-DATC-DATS)%(2*np.pi)
    DNU = DATC-DATS#
    DCHOM = np.cos(DOMEGA*0.5) 
    DSHOM = np.sin(DOMEGA*0.5) 
    DSOM2 = DSOMEG*DSOMEG
    DCHOM2 = DCHOM*DCHOM 
    DSHOM2 = DSHOM*DSHOM 
    DCHIKL = np.cos(DIKL*0.5)  
    DCHIK4 = DCHIKL*DCHIKL*DCHIKL*DCHIKL
    D132S2 = 1-3./2.*DSIKL*DSIKL 
    DMOF65 = (2./3.-DSOM2)*D132S2  
    DMOF66 = DSOM2*DCHIK4
    DMOF67 = DSOMEG*DCHOM2*DCHIK4
    DMOF68 = np.sin(DOMEGA+DOMEGA)*D132S2
    DMOF69 = DSOMEG*DSHOM2*DCHIK4
    DMOF70 = DCHOM2*DCHOM2*DCHIK4
    DMOF71 = DSOM2*D132S2
    DSIH = np.sin(DIH)#
    DCHIH = np.cos(DIH*0.5)
    DSHIH = np.sin(DIH*0.5)
    DS2IH = np.sin(DIH+DIH)
    DTHIH2 = DSHIH**2/DCHIH**2
    DIGHI2 = np.cos(DIH)/DCHIH**2  
    DPMKSI = P_rad-DKSI
    DC2PMK = np.cos(DPMKSI+DPMKSI)
    DQA = 1./np.sqrt(0.25+1.5*DIGHI2*DC2PMK+2.25*DIGHI2*DIGHI2)
    DRA = 1./np.sqrt(1-12*DTHIH2*DC2PMK+36*DTHIH2*DTHIH2)
    
    #actual values used in the component list below (12 uniques) Equations 73 to 79 from Schureman
    DND73 = (2./3.-DSIH**2)/DMOF65
    DND74 = DSIH**2/DMOF66 
    DND75 = DSIH*DCHIH**2/DMOF67 #DFQ1/DFO1
    DND76 = DS2IH/DMOF68 #DFJ1
    DND77 = DSIH*DSHIH**2/DMOF69 
    DND78 = DCHIH**4/DMOF70 #DFN2/DFNU2/DFM2/DFLAB2
    DND79 = DSIH**2/DMOF71 
    DFM1 = DND75/DQA #DFO1/DQA #tugo tides.def points to Schureman eq 144
    DFK1 = np.sqrt(DC5023*DC5023*DS2IH*DS2IH+2*DC5023*DC1681*DS2IH*np.cos(DNU)+DC1681*DC1681)/(DC5023*DMOF68+DC1681) #tugo tides.def points to Schureman eq 227
    DFL2 = DND78/DRA #DFM2/DRA #tugo tides.def points to Schureman eq 215
    DFK2 = np.sqrt(DC5023*DC5023*DSIH**2*DSIH**2+2*DC5023*DC0365*DSIH**2*np.cos(DNU+DNU)+DC0365*DC0365)/(DC5023*DMOF71+DC0365) #tugo tides.def points to Schureman eq 235
    DFM1C = (1-10*DSHIH**2+15*DSHIH**2*DSHIH**2)*DCHIH**2/((1-10*DSHOM2+15*DSHOM2*DSHOM2)*DCHOM2)
    
    multiply_variables = np.stack([DND73,DND74,DND75,DND76,DND77,DND78,DND79,DFM1,DFK1,DFL2,DFK2,DFM1C])
    sel_cols = ['DND73','DND74','DND75','DND76','DND77','DND78','DND79','DFM1','DFK1','DFL2','DFK2','DFM1C']
    
    v0uf_allT = get_schureman_table()
    v0uf_sel = v0uf_allT.loc[const_list]
    f_i = np.ones(shape=(len(const_list), len(dood_date)))
    for variable, colname in zip(multiply_variables, sel_cols): #this loop is faster than array multiplications, since it only calculates the necessary factors (saves a lot of overhead)
        power = v0uf_sel[colname].values
        idnozero = power!=0
        f_i[idnozero,:] *= variable**power[idnozero][np.newaxis].T
    
    v0uf_M2 = v0uf_allT.loc[['M2']]
    f_i_M2 = np.ones(shape=(len(['M2']), len(dood_date)))
    for variable, colname in zip(multiply_variables, sel_cols): 
        power = v0uf_M2[colname].values
        idnozero = power!=0
        f_i_M2[idnozero,:] *= variable**power[idnozero][np.newaxis].T
    
    #create dataframe
    f_i_pd = pd.DataFrame(f_i, index=const_list)
    f_i_M2_pd = pd.DataFrame(f_i_M2, index = ['M2'])
    
    if xfac: #if variable is not: None, False, 0, more?
        f_i_pd = correct_fwith_xfac(f_i_pd, f_i_M2_pd, xfac=xfac)
    
    return f_i_pd


def correct_fwith_xfac(f_i_pd, f_i_M2_pd, xfac):
    """
    Correct f-values with xfactor, this definition is only ran when xfac=True.

    Parameters
    ----------
    f_i_pd : TYPE
        DESCRIPTION.
    f_i_M2_pd : TYPE
        DESCRIPTION.

    Returns
    -------
    f_i_pd : TYPE
        DESCRIPTION.

    """
    
    if isinstance(xfac,dict):
        logger.info(f'xfac dictionary provided: {xfac}')
        xfac_values = xfac
    else: #use default xfactor values
        xfac_values = {'MU2': 0.00, # 0 betekent knoopfactor uit, 1 is helemaal aan
                       'N2':  0.00,
                       'NU2': 0.80,
                       'M2':  0.53,
                       '2MN2':0.20,
                       'S2': -0.82,
                       'M4':  0.70,
                       'MS4': 0.00,
                       'M6':  0.75,
                       '2MS6':0.20,
                       'M8':  0.70,
                       '3MS8':0.60}
    
    for xfac_const in xfac_values.keys():
        if not xfac_const in f_i_pd.index:
            continue
        if all(f_i_pd.loc[xfac_const] == 1): # HET IS EEN NIET-KNOOPAFHANKELIJKE COMPONENT (F=1) (like S2)
            f_i_pd.loc[[xfac_const]] = xfac_values[xfac_const]*(f_i_M2_pd.loc[['M2'],:].values-1)+1 # # hvufea.f line 176. uit v0uf_M2 ipv v0uf_sel, want daar is M2 waarde nog niet gecorrigeerd met xfac (kan ook uit f_i_HAT komen maar daar zit M2 niet per definitie in)
        else: # HET IS EEN KNOOPAFHANKELIJKE COMPONENT (F#1)
            f_i_pd.loc[[xfac_const]] = xfac_values[xfac_const]*(f_i_pd.loc[[xfac_const],:]-1)+1 # uit document tom bogaard, en hvufea.f line 181. staat ook in hatyan gebruikershandleiding bij knoopfactoren (pag2-5)

    return f_i_pd



